# -*- coding: utf-8 -*-
"""
Created on Wed Mar 31 09:58:03 2021

@author: yanni
"""
                    
import snappy  
from snappy import GPF 
from snappy import ProductIO
from snappy import jpy
import time 
import datetime
import os
import rasterio
from rasterio.plot import show
from matplotlib import pyplot
import zipfile


# adjust footprint 
PATH_DATA = '/tmp/ndsi/data'
PATH_OUT = '/tmp/ndsi/out'
PATH_NDVI = '/tmp/ndsi/ndvi'
PATH_NDSI = '/tmp/ndsi/ndsi'

JPY_BAND_STRING = "org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor"

CONFIG = {
        "preprocessing": [
            {
                "step_name": 'Water_Mask',
                "param_type": 'float32',
                "preprocessing_expression": 'scl_water'},
            {
                "step_name": 'Cloud_Mask',
                "param_type": 'float32',
                "preprocessing_expression": 'quality_cloud_confidence'},
            {
                "step_name": 'CloudMask',
                "param_type": 'float32',
                "preprocessing_expression": 'if Cloud_Mask > 90 then 2 else if Cloud_Mask > 50 AND B8 > 0.3 then 1 else 0'},
            {
                "step_name": 'NDSI',
                "param_type": 'float32',
                "preprocessing_expression": '(B3-B11)/(B3+B11)'},
            {
                "step_name": 'NDVI',
                "param_type": 'float32',
                "preprocessing_expression": '(B8-B4)/(B8+B4)'},
            {
                "step_name": 'snow_pass_1',
                "param_type": 'float32',
                "preprocessing_expression": 'if CloudMask < 2 AND NDSI > 0.4 AND B4 > 0.2 AND NDVI < 0 AND Water_Mask == 0 then 1 else 0'},
            {
                "step_name": 'snowmask',
                "param_type": 'float32',
                "preprocessing_expression": 'if CloudMask < 2 AND NDSI > 0.150 AND B4 > 0.04 AND NDVI < 0 AND Water_Mask == 0 OR snow_pass_1 == 1 then 1 else if CloudMask > 0 then 9 else 0'},
            {
                "step_name": 'snow_cover',
                "param_type": 'float32',
                "preprocessing_expression": 'if snowmask == 1 then 1 else 0'},
            ]
        }

# init Snappy
jpy = snappy.jpy
BandDescriptor = jpy.get_type('org.esa.snap.core.gpf.common.BandMathsOp$BandDescriptor')


def do_resampling(source):
    # Resamples all the bands to 30x30 m (B3)
    print('Resampling')
    parameters = snappy.HashMap()
    parameters.put('Resample', True)
    parameters.put('referenceBand', 'B3') # basically we use B3
    parameters.put('upsampling', 'Nearest')
    parameters.put('downsampling', 'First')
    parameters.put('flagDownsampling', 'First')
    output = snappy.GPF.createProduct('Resample', parameters, source)
    return output

def do_subset(source):
    # Defines the research area over Zugspitze 
    print('Subset')
    parameters = snappy.HashMap()
    parameters.put('copyMetadata', True)
    # adjust footprint
    footprint = "POLYGON((13.278350830078125 52.453917235802294,13.278350830078125 52.586780621860804,13.521423339843752 52.586780621860804,13.521423339843752 52.453917235802294,13.278350830078125 52.453917235802294))"
    parameters.put('geoRegion', footprint)
    sBands = "B1,B2,B3,B4,B8,B11,B12"
    parameters.put('sourceBands', sBands )
    output = snappy.GPF.createProduct('Subset', parameters, source)    
    return output

def do_BandMerge(master, slave):
    # merges the last two products 
    print('Band Merge')
    sourceProducts= snappy.HashMap()
    sourceProducts.put('masterProduct', master)
    sourceProducts.put('slaveProduct', slave)
    parameters = snappy.HashMap()
    output = snappy.GPF.createProduct('Merge', parameters, sourceProducts)
    return output

def apply_preprocessing(source_data, step_name, param_type, preprocessing_expression):
    snappy_parameters = snappy.HashMap()

    preprocessing_step = BandDescriptor()
    preprocessing_step.name = step_name
    preprocessing_step.type = param_type
    preprocessing_step.expression = (preprocessing_expression)

    process_request = jpy.array(JPY_BAND_STRING, 1)
    process_request[0] = preprocessing_step

    snappy_parameters.put('targetBands', process_request)
    processed_data = snappy.GPF.createProduct('BandMaths', snappy_parameters, source_data)

    return do_BandMerge(source_data, processed_data)



def main(path_data, path_out, path_ndsi, path_ndvi, config):
    for zip_file in os.listdir(path_data):
        if zip_file.endswith('.zip'):
            with zipfile.ZipFile(path_data + zip_file, 'r') as zip_ref:
                zip_ref.extractall(path_data)
            os.remove(path_data + zip_file)

    if not os.path.exists(path_out):
        os.makedirs(path_out)

    if not os.path.exists(path_ndvi):
        os.makedirs(path_ndvi)

    if not os.path.exists(path_ndsi):
        os.makedirs(path_ndsi)


    for files in os.listdir(path_data):
        snowcover_filepath = path_out + files + '_snowcover.tif'

        if os.path.exists(snowcover_filepath):
            print("snow cover output file already exists")
        else:
            loopstarttime=str(datetime.datetime.now())
            print('Start time:', loopstarttime)
            start_time = time.time()
            print('reading file...')

            sentinel_2 = snappy.ProductIO.readProduct(path_data + files)

            # Data Preparation
            aligend_data = do_resampling(sentinel_2)
            trimmed_data = do_subset(aligend_data)

            current_data_object = trimmed_data
            for step in config["preprocessing"]:
                step["source_data"] = current_data_object
                current_data_object = apply_preprocessing(**step)

            # Data Export
            snappy.ProductIO.writeProduct(current_data_object, path_out + files + '_snowcover', 'GeoTIFF')

            sentinel_2.dispose()
            sentinel_2.closeIO()  

# <--- Disabled Block
#   This block does not work with the current gdal installation
#   We need to fix it  before we try this out!
#
#    # Evaluation
#    ## Snowcover Calculatio
#    for filename in os.listdir(path_out):
#        if filename.endswith('.tif'):
#            filepath = os.path.join(path_out,filename)
#            ds = gdal.Open(filepath)
#            stats = ds.GetRasterBand(1).GetStatistics(1,1)
#            snowcover_percentage = stats[2]*100
#            print(snowcover_percentage)
# <--- Disabled Block

    # Visulation
    ## NDSI Visulation
    for file in os.listdir(path_ndsi): 
        day = file[17:19]
        month = file[15:17]
        year = file[11:15]
        filepath = path_ndsi + file
        src = rasterio.open(filepath)
        title_ndsi = ('NDSI for %s - %s - %s' %(day,month,year))
        show(src.read(1), transform=src.transform, cmap='gray', title=title_ndsi)
        pyplot.show()

    ## NDVI Calculatio
    for file in os.listdir(path_ndvi):
        day = file[17:19]
        month = file[15:17]
        year = file[11:15]
        filepath = path_ndvi + file
        src = rasterio.open(filepath)
        title_ndvi = ('NDVI for %s - %s - %s' %(day,month,year))
        show(src.read(1), transform=src.transform, cmap='Greens', title=title_ndvi)
        pyplot.show()

    print('done')

if __name__== "__main__":
    main(PATH_DATA, PATH_OUT, PATH_NDSI, PATH_NDVI, CONFIG)
